# Details on check matrix operations

## Basic features of check matrix

- All action on state can be represented as operations on check matrix.
- QuantumLegos which don't connect each other are represented by block diagonal check matrix.

Thus, one needs only one check matrix to represent state.

- Generated group from generators represented by check matrix is invariant under row-swap and row-multiplying transforms on the check matrix.

Thus, one can freely perform these row operations on check matrices.

## Construction of check matrices and retrieving features

Use [`CheckMatrix`](@ref) from `Matrix{Bool}` and [`checkmatrix`](@ref) from `Vector{PauliOp}`.

```@docs
CheckMatrix
checkmatrix
```

Get xpart and zpart of check matrix.

```@docs
QuantumLegos.xpart
QuantumLegos.zpart
```

Get generators from check matrix with [`generators`](@ref).

```@docs
generators
```

See also, [`GeneratedPauliGroup`](@ref) and [`Base.Set`](https://docs.julialang.org/en/v1/base/collections/#Base.Set) to get group generated by generators.

## Direct sum

Used when adding lego without edge.

```@docs
QuantumLegos.direct_sum
```

## Self-tracing

### self-tracing

```@docs
QuantumLegos.self_trace!
```

During self-tracing, pre-formatting and post-formatting described below is performed.

### pre-formatting

```@docs
QuantumLegos.eliminate_column!
```

```@docs
QuantumLegos.align_row!
```

[`QuantumLegos.align_row!`](@ref)

By this process, all columns to be traced have `1`s on only 1st to 4th rows.
So during [`QuantumLegos.self_trace!`](@ref), all stabilizers generated from top three rows are calculated and perform operator matching.

### post-formatting

Remove rows which are linear combinations of other rows.

```@docs
QuantumLegos.ref!
```

```@docs
QuantumLegos.eliminate_dependent_row!
```

